from matplotlib import patches
import numpy as np

### ATTRIBUTE MAP
### for each patch, a dict of attribute, and the required function to change that attribute
ATTRIBUTE_MAP = {
	"Circle" : dict( xy = "set_center", radius = "set_radius")

}

class AnimatedPatch():
	"""Object for animating the input patch."""

	def __init__(self, patch_type, *patch_args, patch_id=0, animate_kwargs = None, n_frames=0, **patch_kwargs):
		"""
		:param patch_type: string corresponding to type of patch.
		:param animate_kwargs: dict of patch_attr : numpy array of values
		:param patch_kwargs: kwargs to be passed to patch
		"""

		assert patch_type in dir(patches), f"Patch '{patch_type}' not found in patches."

		self.patch_type = patch_type
		self.patch = getattr(patches, patch_type)(*patch_args, **patch_kwargs)

		self.animate_kwargs = animate_kwargs
		if self.animate_kwargs is not None:
			self.animate = True
			self.length = n_frames
		else:
			self.animate = False
			self.length = 0

		self.id = patch_id
		self.n_frames = n_frames
		self.c = 0

	def __len__(self):
		return self.length

	def __iter__(self):
		return self

	def __next__(self):
		if self.animate_kwargs is None: return None
		if self.c > len(self): raise StopIteration

		for attr, arr in self.animate_kwargs.items():

			frame = int(np.ceil((self.c / self.n_frames) * (len(arr) - 1))) # get relevant frame sample of this array (sample at axis fps)
			val = arr[frame]

			setter = getattr(self.patch, ATTRIBUTE_MAP[self.patch_type][attr]) # function to set value
			setter(val)

		self.c += 1

	def get_bounds(self):
		"""Get all x_bounds, y_bounds for patch"""
		patch = self.patch
		if self.patch_type == "Circle":
			x, y = patch._center
			r = patch.get_radius()
			if not self.animate:
				return [[x - r, x + r], [y - r, y + r]]
			else:
				x, y = (x,y) if "xy" not in self.animate_kwargs else np.swapaxes(self.animate_kwargs["xy"], 0, 1)
				r = r if "radius" not in self.animate_kwargs else self.animate_kwargs["radius"]
				return [ [(x - r).min(), (x + r).max()], [(y-r).min(), (y+r).max()] ]


def animate_patch(obj, *args, **kwargs):

	pass #


from matplotlib import pyplot as plt
from figure import FuncFig
import numpy as np

if __name__ == "__main__":
	fig, ax = plt.subplots(FigureClass=FuncFig, animate=True, time = 5)

	# f1 = ax.add_func(lambda x: np.exp(- (x-0.5) ** 2 / 0.1), x0=0, x1 = 1, marker="s", markevery=[5, 20, 30], timings=[0,1])
	# f2 = ax.add_func(lambda x: np.sin(5*x), x0=0, x1=1, sample_per_it=2, color="blue", timings=[2, 3])
	# ax.fill_between(f1, f2, timings=[2,3])

	f = ax.add_func(lambda x, t: np.exp( - (x - 0.5) ** 2 / (0.01*t+1e-3)**.5), x0=0, x1=1, timings=[0, 5], is_functional=True)

	fig.render()

	# fig.save()

	plt.show()



"""Introduces the Function class, a class which represents a mathematical function,
which can be plotted as an animation, or visualised as a graph"""

import numpy as np
from matplotlib.collections import PolyCollection


plot_defaults = {
	"lw" : 3,
	"color" : "black",
}

def add_point_to_line(l, x, y):
	"""Add (x,y) point to pyplot line2d"""
	l.set_xdata(np.concatenate([l._x, [x]]))
	l.set_ydata(np.concatenate([l._y, [y]]))

def add_points_to_line(l, xs, ys):
	"""Add points to pyplot line2d"""
	l.set_xdata(np.concatenate([l._x, xs]))
	l.set_ydata(np.concatenate([l._y, ys]))

class Function:

	def __init__(self, func, xrange=None, animate=False, id=0, **plt_kw):

		self.id = id
		self.func = func

		if xrange is None: xrange = np.arange(0, 1, 0.05)

		self.x = xrange
		self.plot = None # Line2D object

		self.plt_kw = plot_defaults.copy()
		# replace default plot keywords with any given
		for kw, val in plt_kw.items():
			self.plt_kw[kw] = val

	def set_plot(self, plot):
		self.plot = plot

	def __call__(self, *args, **kwargs):
		return self.func(*args, **kwargs)

class XSampler():
	"""Used to sample a numpy xrange in order to effectively provide smooth animations"""
	def __init__(self, x0=0, x1=1, n_samples=10, sample="linear"):
		self.x = np.linspace(x0, x1, n_samples)

class AnimatedFunction(Function):
	"""Subclass of function, used to produce animations"""

	def __init__(self, x0=0, x1=1, x_samples=10, sampling="linear", sample_per_it = 1, **kwargs):
		"""
		:param sample_data: tuple of (x_start, x_end, n_samples)
		:param sampling:
		:param kwargs:
		"""
		super().__init__(**kwargs)
		self.x = XSampler(x0, x1, x_samples * sample_per_it, sampling).x

		self.sample_per_it = sample_per_it
		self.c = 0 # counter

	def set_plot(self, plot):
		"""Take Line2D object, save to function"""
		# For animating, mark every param needs to be stored separately to render through animation
		if plot._markevery is not None:
			self.markevery = np.array(plot._markevery)
			plot._markevery = []
		else:
			self.markevery = None

		return super().set_plot(plot)

	def __iter__(self): return self

	def __next__(self):
		xs, ys = [], []
		for n in range(self.sample_per_it):
			if self.c >= len(self):
				add_points_to_line(self.plot, xs, ys) # add points so far
				raise StopIteration

			if self.markevery is not None:
				self.plot._markevery = self.markevery[self.markevery < self.c] # update markevery

			x = self.x[self.c]
			y = self(x)

			xs.append(self.x[self.c])
			ys.append(y)

			self.c += 1

		add_points_to_line(self.plot, xs, ys)

	def __len__(self): return len(self.x)

class AnimatedFunctional(Function):
	"""A univariable function f(x), which changes with t. """

	def __init__(self, func,  t0=0, t1=1, t_samples=10, t_sampling="linear", x0=0, x1=1, x_samples = 100, xrange=None, *args, **kwargs):

		if xrange is None:
			xrange = XSampler(x0, x1, x_samples).x

		self.t = XSampler(t0, t1, t_samples, t_sampling).x
		self.functional = func

		self.c = 0
		super().__init__(lambda x: func(x, t=t0), xrange=xrange, *args, **kwargs)

	def __next__(self):
		if self.c >= len(self): raise StopIteration
		t = self.t[self.c]
		self.func = lambda x: self.functional(x, t=t)

		self.plot.set_xdata(self.x)
		self.plot.set_ydata(self.func(self.x))

		self.c += 1

	def __iter__(self): return self
	def __len__(self): return len(self.t)

class FillBetween(PolyCollection):
	"""Fills between two functions. Can be animated."""

	def __init__(self, ax, func1, func2, animate=False, obj_id=0, **fill_kwargs):
		self.ax = ax
		self.id = obj_id
		self.func1, self.func2 = func1, func2
		self.animate = animate
		if not np.array_equal(func1.x, func2.x):
			print("Func1.x and func2.x do not match for fill_between. Using func1.x")

		self.x = x = func1.x

		if not animate:
			ax.fill_between(x, func1(x), func2(x), **fill_kwargs)

		self.fill_kwargs = fill_kwargs
		self.c = 0 # counter

	def get_animated(self):
		return self.animate

	def __len__(self):
		return len(self.x) - 1 # only go to penultimate x val

	def __iter__(self): return self

	def __next__(self):

		if self.c >= len(self): raise StopIteration

		x0, x1 = self.x[self.c:self.c+2]
		x1+=(x1-x0)*0.5 # padding to prevent whitespace
		f1, f2 = self.func1, self.func2
		verts = np.array([[[x, f(x)] for x, f in [[x0, f1], [x0, f2], [x1, f2], [x1, f1] ]]])

		collection = PolyCollection(verts, **self.fill_kwargs)
		self.ax.add_collection(collection, autolim=False)

		self.c += 1


import os

def try_mkdir(dir):
	"""If dir does not exist, make it"""

	if os.path.isdir(dir):
		return None

	print(f"Making directory: {dir}")
	os.mkdir(dir)

np_minmax = lambda x: [x.min(), x.max()]



import matplotlib.projections as proj

from matplotlib.animation import FuncAnimation, writers
from matplotlib import pyplot as plt
from tqdm import tqdm

import math
import numpy as np
from viseng.utils import *
from viseng.function import Function, AnimatedFunction, FillBetween, AnimatedFunctional
from viseng.annotations import TextBox
from viseng.patches import AnimatedPatch

def load_writer(writer_name, message=""):

	try:
		writer = writers[writer_name]

	except RuntimeError as e:
		print(f"Runtime Error: {e}")
		if message != "": print(message)
		return None

	return writer

class AnimTiming(dict):
	"""Object for handling animation timing.
	Dictionary of 'id' : [time start: time end]"""

	def __init__(self, **kwargs):
		super().__init__(**kwargs)

class FuncAxis(plt.Axes):
	"""Axis for visualising functions"""
	name = "FuncAxis"

	def __init__(self, *args, **kwargs):

		fig = self.figure
		self.fps = fig.fps
		self.anim_timings = fig.anim_timings

		self.functions = [] # list of Function objects
		self.other_objects = [] # list of Patch objects to iterate through (such as fill_between)

		self.animate = fig.animate
		self.anim = None # anim object

		super().__init__(rect=args[1], fig=fig, **kwargs) # initialise regular axis

		# defaults, modified by functions entered
		self.set_xlim(0, 1)
		self.set_ylim(0, 1)

		self.id_counter = 0

	def extend_limits(self, x=None, y=None):
		"""For all given x and y values, extend axis limits to incorporate those values"""
		items = [ [x, self.get_xlim(), self.set_xlim], [y, self.get_ylim(), self.set_ylim] ]

		for data, lims, set in items:
			if data is None: pass
			lower = min(*lims, *data)
			upper = max(*lims, *data)
			set(lower, upper)

	def new_id(self):
		self.id_counter +=1
		return self.id_counter - 1

	def add_func(self, function, timings=None, is_functional=False, **kwargs):
		"""Add new function to axis.

		if timings provided, array of [start, end]. Otherwise, assumed to be entire animation [start, end]"""

		func_id = self.new_id()

		if "animate" not in kwargs: kwargs["animate"] = self.animate # animate attribute, default to axis flag
		animate = kwargs["animate"]
		if animate:
			timings = self.anim_timings["fig"] if timings is None else timings
			self.anim_timings[func_id] = timings

			if not is_functional:
				if "x_samples" not in kwargs: kwargs["x_samples"] = int(math.ceil((timings[1] - timings[0]) * self.fps))
				func = AnimatedFunction(func=function, id=func_id, **kwargs)

			else:
				if "t_samples" not in kwargs: kwargs["t_samples"] = int(math.ceil((timings[1] - timings[0]) * self.fps))
				func = AnimatedFunctional(func=function, id=func_id, **kwargs)

		else:
			func = Function(function, id=func_id, **kwargs)
		self.functions.append(func)

		# Update axis bounds
		self.extend_limits(x=np_minmax(func.x), y=np_minmax(func(func.x)))

		return func

	def static_plot(self):
		for f in self.functions:
			f.plot, = self.plot(f.x, f(f.x), **f.plt_kw)

	def update_anim(self):
		for f in self.functions:
			f.set_plot(self.plot([], [], **f.plt_kw)[0])

		all_objects = [o for o in [*self.functions, *self.other_objects]]
		all_iters = [iter(o) for o in all_objects]

		for t in np.arange(0, self.figure.time, 1/self.figure.fps):
			for n, o in enumerate(all_iters):
				t_start, t_end = self.anim_timings[all_objects[n].id]
				if t_start <= t < t_end:
					try: next(o)
					except StopIteration: pass

			yield None

		# for plots in zip_longest(*self.functions, *self.others):
		# 	yield None	# yield at end of every frame to wait for next drawing

	def animated_plot(self):
		update = self.update_anim()
		def anim(i):
			"""Render next frame"""
			try: next(update)
			except StopIteration: return None

		return anim

	def render(self, *args, **kwargs):
		if self.animate: return self.animated_plot()
		else: self.static_plot()

	def fill_between(self, func1, func2, timings=None, **kwargs):
		obj_id = self.new_id()
		animate = self.animate if "animate" not in kwargs else kwargs['animate'] # animate attribute, default to axis flag
		if animate:
			self.anim_timings[obj_id] = self.anim_timings["fig"] if timings is None else timings
			fill = FillBetween(self, func1, func2, animate=animate, obj_id=obj_id, **kwargs)
			self.other_objects.append(fill)

	def add_text(self, s, x, y, **text_kwargs):
		text_id = self.new_id()
		text = TextBox(self, s, x, y, text_id=text_id, **text_kwargs)
		self.anim_timings[text_id] = self.anim_timings["fig"]  # timings from main figure
		self.other_objects.append(text)

	def add_patch(self, patch_type, *patch_args, animate_kwargs = None, timings=None, **patch_kwargs):
		patch_id = self.new_id()
		self.anim_timings[patch_id] = self.anim_timings["fig"] if timings is None else timings
		# get number of frames of sequence (take off one to deal with rounding errors, ensures all clip is in time
		n_frames = int((self.anim_timings[patch_id][1] - self.anim_timings[patch_id][0]) * self.fps) - 1

		patch = AnimatedPatch(patch_type, *patch_args, animate_kwargs=animate_kwargs, n_frames=n_frames,
							  patch_id = patch_id, **patch_kwargs)
		self.other_objects.append(patch)
		super().add_patch(patch.patch)  # add patch to axes

		self.extend_limits(*patch.get_bounds())  # update bounds for frame=0

proj.register_projection(FuncAxis)

class FuncFig(plt.Figure):
	"""Used to generate custom axes"""

	def __init__(self, *args, animate=False, fps=30, time = 5, **kwargs):
		super().__init__(*args, **kwargs)
		self.animate = animate
		self.fps = fps
		self.time = time
		self.n_frames = fps * time

		self.anim_timings = AnimTiming(fig = [0, time])

	def add_subplot(self, *args, **kwargs):
		"""Add FuncAxis as subplot"""
		return super(FuncFig, self).add_subplot(*args, **kwargs, projection="FuncAxis")

	def render(self):

		funcs = [ax.render() for ax in self.axes]

		def anim(i):
			[f(i) for f in funcs]

		self.anim = FuncAnimation(fig=self, func=anim, frames=self.n_frames, interval = 1000//self.fps) # attribute func animation to permanent object

	def save(self, *args, **kwargs):

		if self.animate:
			self.saveanim(*args, **kwargs)

		else:
			super().savefig(*args, **kwargs)

	def saveanim(self, title="unnamed", fmt="mp4", out_dir = "outputs"):

		# writer = writers["xxx"]
		if fmt == "gif":
			writer = load_writer("imagemagick",
								 message="imagemagick not found on system. Required for gifs - check installed and try again")

		elif fmt == "mp4":
			writer = load_writer("ffmpeg")

		else:
			print(f"Format '{fmt}' not currently supported. Only 'mp4', 'gif' accepted.")
			return None

		w = writer(fps = self.fps, bitrate = 1500)

		try_mkdir(out_dir)

		with tqdm(total = self.n_frames) as save_progress:
			self.anim.save(os.path.join(out_dir, f"{title}.{fmt}"), writer=w,
						   progress_callback = lambda x,i: save_progress.update())

	def extend(self, left=0, right=1, bottom=0, top=1):
		"""Extend subplot. By default, full size"""
		self.subplots_adjust(left=left, right=right, bottom=bottom, top=top)


from matplotlib import pyplot as plt
from matplotlib.text import Text

class TextBox(Text):

	def __init__(self, ax, s="Text", x=0, y=0, strings = None, text_id=None, **text_kwargs):
		"""
		:param ax:
		:param s:
		:param x:
		:param y:
		:param strings: list of strings, used for animating text
		:param text_kwargs:
		"""
		super().__init__(x=x, y=y, **text_kwargs)

		self.id = text_id
		self.strings = strings
		self.c = 0 # counter
		ax.add_artist(self)

	def __iter__(self):
		if self.strings is None:
			raise ValueError("Text cannot be made animated - no strings given")

		return self

	def __next__(self):
		if self.c >= len(self): raise StopIteration

		self.set_text(self.strings[self.c])

		self.c += 1

	def __len__(self):
		"""Length if animated, in number of provided strings"""
		return len(self.strings)


from setuptools import setup

setup(
	name='viseng',
	version='0.1.1',
	packages=['viseng'],
	url='https://github.com/OllieBoyne/viseng',
	license='',
	author='ollie',
	author_email='ollieboyne@gmail.com',
	description='Visualisation library for producing graphics from Python\'s Matplotlib'
)
